//! This module contains everything related to rendering.

pub mod camera;
pub mod controls;

use glium::draw_parameters::DepthTest;
use glium::texture::{
    RawImage2d,
    SrgbTexture2d,
};
use glium::index::{
    NoIndices,
    PrimitiveType
};
use glium::{
    Program,
    Display,
    VertexBuffer,
    Frame,
    DrawParameters,
    Depth,
};

use glium::backend::Facade;
use glium::backend::glutin::headless::Headless;

use image;

use model::{Model, Vertex};
use renderer::camera::RenderCamera;

/// Anything on which you can call draw to get a frame.
pub trait Drawer {
    /// Get a frame from the drawer.
    fn draw(&self) -> Frame;
}

impl Drawer for Display {
    fn draw(&self) -> Frame {
        Display::draw(self)
    }
}

impl Drawer for Headless {
    fn draw(&self) -> Frame {
        Headless::draw(self)
    }
}

/// A material that can be bound to OpenGL.
///
/// Only textures are supported for now.
pub struct RenderMaterial {
    texture: Option<SrgbTexture2d>,
}

impl RenderMaterial {

    /// Creates a new material with no texture.
    pub fn new() -> RenderMaterial {
        RenderMaterial {
            texture: None
        }
    }

    /// Creates a material from a path to an image file.
    pub fn from_texture_path<D: Drawer + Facade + Sized>(path: &str, drawer: &D) -> RenderMaterial {
        let image = image::open(path);

        if let Ok(image) = image {
            let image = image.to_rgba();
            let dim = image.dimensions();
            let image = RawImage2d::from_raw_rgba_reversed(&image.into_raw(), dim);
            RenderMaterial {
                texture: SrgbTexture2d::new(drawer, image).ok()
            }
        } else {
            RenderMaterial {
                texture: None
            }
        }
    }
}

/// A renderer. It contains a display, shaders, and a vector of models it can render.
pub struct Renderer<'a, D: Drawer + Facade> {
    drawer: D,
    program: Program,
    models: Vec<(&'a Model, Vec<(RenderMaterial, VertexBuffer<Vertex>)>)>,
}

impl<'a, D: Drawer + Facade + Sized> Renderer<'a, D> {
    /// Creates a new renderer from a display.
    ///
    /// Is uses the default shaders and creates an empty vec of models.
    pub fn new(drawer: D) -> Renderer<'a, D> {

        let program = Program::from_source(
            &drawer,
            include_str!("../../assets/shaders/shader.vert"),
            include_str!("../../assets/shaders/shader.frag"),
            None
        );

        if let Ok(program) = program {
            Renderer {
                drawer: drawer,
                program: program,
                models: vec![],
            }
        } else {
            println!("{:?}", program.err().unwrap());
            panic!()
        }

    }

    /// Adds a model to the renderer, and compute the corresponding buffers for rendering.
    pub fn add_model(&mut self, model: &'a Model) {
        let mut buffers = vec![];

        for part in &model.parts {

            let material = if let Some(ref material_name) = part.material_name {
                if let Some(material) = model.materials.get(material_name) {
                    if let Some(path) = material.textures.get("map_Kd") {
                        RenderMaterial::from_texture_path(path, &self.drawer)
                    } else {
                        RenderMaterial::new()
                    }
                } else {
                    RenderMaterial::new()
                }
            } else {
                RenderMaterial::new()
            };

            let shape = part.build_shape(&model.vertices, &model.texture_coordinates, &model.normals);
            buffers.push((material, VertexBuffer::new(&self.drawer, &shape).unwrap()));
        }

        self.models.push((model, buffers));
    }

    /// Returns the reference to the drawer used for this renderer.
    pub fn facade(&self) -> &D {
        &self.drawer
    }

    /// Creates a frame from the display.
    pub fn draw(&self) -> Frame {
        self.drawer.draw()
    }

    /// Renders the result of the models from the camera and paint it on the target.
    pub fn render<C: RenderCamera>(&self, camera: &C, target: &mut Frame) {
        use glium::Surface;
        target.clear_color_and_depth((0.0, 0.0, 0.0, 1.0), 1.0);

        let params = DrawParameters {
            depth: Depth {
                test: DepthTest::IfLess,
                write: true,
                .. Default::default()
            },
            .. Default::default()
        };

        for &(_, ref buffers) in &self.models {

            for &(ref material, ref buffer) in buffers {

                let perspective = camera.get_perspective_matrix();
                let view = camera.get_view_matrix();

                if let &Some(ref texture) = &material.texture {
                    target.draw(
                        buffer,
                        NoIndices(PrimitiveType::TrianglesList),
                        &self.program,
                        &uniform!(
                            tex: texture,
                            perspective: perspective,
                            view: view,
                        ),
                        &params,
                    ).unwrap();
                } else {
                    target.draw(
                        buffer,
                        NoIndices(PrimitiveType::TrianglesList),
                        &self.program,
                        &uniform!(
                            perspective: perspective,
                            view: view,
                        ),
                        &params,
                    ).unwrap();
                };
            }
        }

    }
}

impl<'a> Renderer<'a, Display> {
    /// Shows the window if it was hidden.
    pub fn show(&mut self) {
        self.drawer.gl_window().show();
    }

    /// Hides the window if it was visible.
    pub fn hide(&mut self) {
        self.drawer.gl_window().hide();
    }

}
